/* ... */
/* this file is part of netraf project */
/* Copyright (c) 2005 M.S. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <math.h>
#include "minigui.h"


#define MATRIX_START_LINE 0.0
#define MATRIX_END_LINE (float)(LINES)

#define NR_OF_MATRIX_COLORS 4

#define GREEN_TRAIL_PROBABILITY 84
#define WHITE_TRAIL_PROBABILITY 10
#define CYAN_TRAIL_PROBABILITY 5
#define BLUE_TRAIL_PROBABILITY 1

#define GREEN_HEAD_PROBABILITY 49
#define WHITE_HEAD_PROBABILITY 40
#define CYAN_HEAD_PROBABILITY 10
#define BLUE_HEAD_PROBABILITY 1


#define GUI_THREAD_INIT_DELAY 100000
#define GUI_THREAD_DONE_DELAY 100000




  typedef struct matrixThread_arg_t{
    pthread_t thread_id;

    int thread_stop;

    unsigned long refresh;
  } matrixThread_arg_t;


  /** structure representing one "falling" code chunk */
  typedef struct codechunk_t{
    /** horizontal position on screen */
    int x;

    /** vertical positions on screen */
    float y, erasing_y;

    /** ncurses COLOR_PAIR(n) code-chunk colors */
    long color, trail_color;

    /** speed of chunk head and tail */
    float falling_speed, erasing_speed;

    /** length of chunk head */
    char head_length;

  } codechunk_t;




/**
  Choose color number from probability range.
*/
int compute_color(char *prob_table, int table_size, unsigned int *seedp){
  int i;
  char p=0, n=0;
  char rnd = rand_r(seedp)%100;

  for(i=0; i<table_size; i++){
    n+=prob_table[i];
    if(p <= rnd  &&  rnd < n)
      return i;
    p+=prob_table[i];
  }  

  return 0;
}




void *matrixThread(void *arg){
  matrixThread_arg_t *matrixThread_arg;
  codechunk_t *codechunk;
  long color_table[NR_OF_MATRIX_COLORS];
  char trail_prob_table[NR_OF_MATRIX_COLORS];
  char head_prob_table[NR_OF_MATRIX_COLORS];
  struct timeval now;
  int i;


  usleep(GUI_THREAD_INIT_DELAY);

  /* initialize "random-seed" */
  gettimeofday(&now, NULL);

  /* apropriate thread argument cast */
  matrixThread_arg = (matrixThread_arg_t*)arg;
  
  /* allocate memory for codechunks */
  codechunk = calloc(COLS, sizeof(codechunk_t));
  
  /* initialize color tables */
  color_table[0] = COLOR_PAIR(10) | A_BOLD;
  trail_prob_table[0] = GREEN_TRAIL_PROBABILITY;
  head_prob_table[0] = GREEN_HEAD_PROBABILITY;

  color_table[1] = COLOR_PAIR(0) | A_BOLD;
  trail_prob_table[1] = WHITE_TRAIL_PROBABILITY;
  head_prob_table[1] = WHITE_HEAD_PROBABILITY;

  color_table[2] = COLOR_PAIR(2) | A_REVERSE | A_BOLD;
  trail_prob_table[2] = CYAN_TRAIL_PROBABILITY;
  head_prob_table[2] = CYAN_HEAD_PROBABILITY;

  color_table[3] = COLOR_PAIR(6) | A_REVERSE | A_BOLD;
  trail_prob_table[3] = BLUE_TRAIL_PROBABILITY;
  head_prob_table[3] = BLUE_HEAD_PROBABILITY;

  /* init chunks */
  for(i=0; i<COLS; i++){
    codechunk[i].x = i;
    
    /* reset y positions */
    codechunk[i].y = MATRIX_START_LINE;
    codechunk[i].erasing_y = MATRIX_START_LINE - (float)(rand_r((unsigned int*)(&now.tv_usec))%LINES);

    /* randomize colors */
    codechunk[i].color = color_table[compute_color(head_prob_table, NR_OF_MATRIX_COLORS, (unsigned int*)(&now.tv_usec))];
    codechunk[i].trail_color = color_table[compute_color(trail_prob_table, NR_OF_MATRIX_COLORS, (unsigned int*)(&now.tv_usec))];

    /* randomize falling and erasing speeds */
    codechunk[i].falling_speed = (float)(rand_r((unsigned int*)(&now.tv_usec))%70 + 30) * 0.01;
    codechunk[i].erasing_speed = codechunk[i].falling_speed - 0.20;
    codechunk[i].head_length = rand_r((unsigned int*)(&now.tv_usec))%12+2;
  }

  /* animate chunks!! */
  for(;!matrixThread_arg->thread_stop;){
    usleep(matrixThread_arg->refresh);

    /* do some computations for every chunk */
    for(i=0; i<COLS; i++){

      /* "erasing tail" */      
      if(codechunk[i].erasing_y >= MATRIX_START_LINE  &&  codechunk[i].erasing_y < MATRIX_END_LINE)
        mvwaddch(stdscr, floor(codechunk[i].erasing_y), codechunk[i].x, ' ');

      /* change attributes (to draw "trail") */
      if(codechunk[i].y-codechunk[i].head_length >= MATRIX_START_LINE  &&
         codechunk[i].y-codechunk[i].head_length  <  MATRIX_END_LINE)
        mvwchgat(stdscr,
	         codechunk[i].y-codechunk[i].head_length,
		 codechunk[i].x,
	         1,
		 codechunk[i].trail_color & ~A_BOLD,
		 0,
		 NULL);

      /* draw head */
      if(codechunk[i].y >= MATRIX_START_LINE  && codechunk[i].y < MATRIX_END_LINE){
        wattron(stdscr, codechunk[i].color);
        mvwaddch(stdscr, floor(codechunk[i].y), codechunk[i].x, rand_r((unsigned int*)(&now.tv_usec))%80+35);
        wattroff(stdscr, codechunk[i].color);
      }
  
      /* compute vertical positions */
      codechunk[i].y += codechunk[i].falling_speed;
      codechunk[i].erasing_y += codechunk[i].erasing_speed;

      /* if codechunk falls, reset it's attributes */
      if(codechunk[i].erasing_y >= MATRIX_END_LINE){
        
        /* reset y positions */
        codechunk[i].y = MATRIX_START_LINE;
        codechunk[i].erasing_y = MATRIX_START_LINE - (float)(rand_r((unsigned int*)(&now.tv_usec))%LINES);

        /* randomize colors */
        codechunk[i].color = color_table[compute_color(head_prob_table, NR_OF_MATRIX_COLORS, (unsigned int*)(&now.tv_usec))];
        codechunk[i].trail_color = color_table[compute_color(trail_prob_table, NR_OF_MATRIX_COLORS, (unsigned int*)(&now.tv_usec))];

        /* randomize falling and erasing speeds */
        codechunk[i].falling_speed = (float)(rand_r((unsigned int*)(&now.tv_usec))%70 + 30) * 0.01;
        codechunk[i].erasing_speed = codechunk[i].falling_speed - 0.20;
        codechunk[i].head_length = rand_r((unsigned int*)(&now.tv_usec))%12+2;
      }
    }

    if(rand_r((unsigned int*)(&now.tv_usec))%60 == 0){
      wattron(stdscr, COLOR_PAIR(0) | A_BOLD);
      mvwprintw(stdscr,
                rand_r((unsigned int*)(&now.tv_usec))%(LINES-3)+2,
		rand_r((unsigned int*)(&now.tv_usec))%(COLS-22),
		"http://netrafd.sf.net/");
      wattroff(stdscr, COLOR_PAIR(0) | A_BOLD);
    }


    /* update screen */
    touchwin(stdscr);
    wrefresh(stdscr);
  }

  /* free codechunks memory */  
  free(codechunk);

  usleep(GUI_THREAD_DONE_DELAY);

  pthread_exit(NULL);
}




int main(int argc, char *argv[]){
  matrixThread_arg_t *matrixThread_arg;


  if(begin_gui_mode() == -1)
    exit(EXIT_FAILURE);

  matrixThread_arg = (matrixThread_arg_t*)calloc(1, sizeof(matrixThread_arg_t));
  matrixThread_arg->refresh = 50000;
  pthread_create(&matrixThread_arg->thread_id, NULL, matrixThread, matrixThread_arg);

  wgetch(stdscr);

  matrixThread_arg->thread_stop = 1;
  pthread_join(matrixThread_arg->thread_id, NULL);
  free(matrixThread_arg);

  end_gui_mode();
  exit(EXIT_SUCCESS);
}
